# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Higher-level utilities for working with tree structures."""

from collections import abc
import functools
from typing import Any, Callable, Generic, Literal, Mapping, NamedTuple, Protocol, TypeVar, overload

import jax
from orbax.checkpoint._src.tree import parts_of
from orbax.checkpoint._src.tree import utils

T = TypeVar('T')
PyTree = Any
PyTreeOf = PyTree | T


class Diff(NamedTuple):
  lhs: Any
  rhs: Any


def tree_difference(
    a: PyTree,
    b: PyTree,
    *,
    is_leaf: Callable[[PyTree], bool] | None = None,
    leaves_equal: Callable[[PyTree, PyTree], bool] | None = None,
):
  """Recursively compute differences in tree structure.

  Lists, tuples, named tuples, mappings and custom PyTree nodes are treated
  as internal nodes and compared element-wise. Other types are treated as
  leaf nodes and compared by type or using a custom equality function, if
  one is provided.

  All leaves of a subtree must be equal according to the equality function to
  avoid being reported as differences.

  The result is a tree that is structurally a prefix of both the inputs.
  Differences in equality of leaf nodes, sequence length or mapping key set at
  corresponding nodes of the two inputs are summarized as `Diff` instances at
  the resulting node in the output.

  Map values that do not differ are omitted from the output; sequence elements
  that do not differ are represented as None.

  For example:
  ```
  first  = {'x': [1, 2, 3], 'y': 'same', 'z': {'p': {}, 'q': 4, 'r': 6}}
  second = {'x': [1, 3],    'y': 'same', 'z': {'p': {}, 'q': 5, 's': 6}}
  tree_difference(first, second) == {
      'x': Diff(lhs=(list, 3), rhs=(list, 2)),
      'z': {'r': Diff(lhs=6, rhs=None), 's': Diff(lhs=None, rhs=6)},
  }
  ```

  Args:
    a: The first object to compare.
    b: The second object to compare.
    is_leaf: Optional function to determine if a node is a leaf. If None,
      defaults to `jax.tree_util.all_leaves`.
    leaves_equal: Optional equality function to apply between corresponding
      leaves of the two data structures. If None, leaf nodes are considered
      equal if they are of the same type.

  Returns:
    A tree summarizing the structural differences between the arguments, or
    `None` if there are no such differences.
  """
  is_leaf = is_leaf or (lambda t: jax.tree_util.all_leaves([t]))
  leaves_equal = leaves_equal or (lambda a, b: type(a) is type(b))

  if is_leaf(a) and is_leaf(b):
    # `a` and `b` are leaf nodes; compare using equality function.
    if leaves_equal(a, b):
      return None
    else:
      return Diff(lhs=a, rhs=b)

  if type(a) is not type(b):
    # Can't check if either node is nested if the node types are different.
    return Diff(lhs=type(a), rhs=type(b))

  if utils.isinstance_of_namedtuple(a):
    diffs = type(a)(*(
        tree_difference(aa, bb, is_leaf=is_leaf, leaves_equal=leaves_equal)
        for aa, bb in zip(a, b)
    ))
    return None if all(d is None for d in diffs) else diffs
  elif isinstance(a, (list, tuple)):
    if len(a) != len(b):
      return Diff(lhs=(type(a), len(a)), rhs=(type(b), len(b)))
    diffs = type(a)(
        tree_difference(aa, bb, is_leaf=is_leaf, leaves_equal=leaves_equal)
        for aa, bb in zip(a, b)
    )
    return None if all(d is None for d in diffs) else diffs
  elif isinstance(a, Mapping):
    diffs = {
        k: diff
        for k, diff in (
            (
                k,
                tree_difference(
                    a[k], b[k], is_leaf=is_leaf, leaves_equal=leaves_equal
                ),
            )
            for k in a
            if k in b
        )
        if diff is not None
    }
    diffs.update((k, Diff(lhs=v, rhs=None)) for k, v in a.items() if k not in b)
    diffs.update((k, Diff(lhs=None, rhs=v)) for k, v in b.items() if k not in a)

    return diffs or None
  else:
    # Custom PyTree nodes. This is limited to the case where `a` and `b` have
    # the same list of children, because `unflatten` requires a `tree_def` that
    # has come from one of them or the other. It's not clear that there's any
    # way to remove that limitation.
    a_keys_and_children, a_tree_def = utils.tree_flatten_with_path_one_level(a)
    b_keys_and_children, _ = utils.tree_flatten_with_path_one_level(b)
    a = {k: v for k, v in a_keys_and_children}
    b = {k: v for k, v in b_keys_and_children}
    diffs = {
        k: diff
        for k, diff in (
            (
                k,
                tree_difference(
                    a[k], b[k], is_leaf=is_leaf, leaves_equal=leaves_equal
                ),
            )
            for k in a
            if k in b
        )
        if diff is not None
    }

    diffs.update((k, Diff(lhs=v, rhs=None)) for k, v in a.items() if k not in b)
    diffs.update((k, Diff(lhs=None, rhs=v)) for k, v in b.items() if k not in a)
    if diffs:
      return jax.tree.unflatten(
          a_tree_def, [diffs.get(k) for k, _ in a_keys_and_children]
      )

    return None


class TrimmedStructureCallback(Protocol, Generic[T]):

  def __call__(
      self,
      path: tuple[str | int, ...],
      structure: PyTreeOf[T],
  ) -> None:
    ...


@overload
def tree_trim(
    template: PyTreeOf[Any],
    structure: PyTreeOf[T],
    *,
    trimmed_structure_callback: TrimmedStructureCallback[T] | None = None,
    strict: Literal[False],
) -> parts_of.PartsOf[PyTreeOf[T]]:
  ...


@overload
def tree_trim(
    template: PyTreeOf[Any],
    structure: PyTreeOf[T],
    *,
    trimmed_structure_callback: TrimmedStructureCallback[T] | None = None,
    strict: Literal[True] = True,
) -> PyTreeOf[T]:
  ...


def tree_trim(
    template: PyTreeOf[Any],
    structure: PyTreeOf[T],
    *,
    trimmed_structure_callback: TrimmedStructureCallback[T] | None = None,
    strict: bool = True,
) -> PyTreeOf[T] | parts_of.PartsOf[PyTreeOf[T]]:
  """Removes nodes in `structure` so that its shape matches that of `template`.

  Only dictionary entries are trimmed; sequences are unchanged and the length
  of a sequence node in `structure` must match that of the corresponding node
  in `template`.

  If `not strict`, any subtree of a mapping or named tuple node of `template`
  that is missing from the corresponding node of `structure` will be replaced
  with an appropriately-shaped subtree full of `...` placeholders (Ellipsis)
  instead of causing an error. In this mode, the tree structure of the result
  is guaranteed to match the tree structure of `template`.

  Args:
    template: The tree whose shape is to be matched.
    structure: The tree to be trimmed.
    trimmed_structure_callback: If present, will be called with the path to, and
      value of, any node that is removed from `structure`.
    strict: Require every element of `template` to be matched by an element of
      `structure`.

  Returns:
    A subset of `structure` that has the same shape as `template`.

  Raises:
    TypeError: If the type of a node in `structure` does not match the
      type of the corresponding node in `template`.
    ValueError: If keys in a dictionary node in `template` are not present
      in the corresponding node in `structure`, or if the length of a sequence
      node in `structure` does not match the length of the corresponding
      sequence node in `template`, or if an internal node that isn't a
      sequence or dictionary is encountered.
  """
  result = _tree_trim_impl(
      template,
      structure,
      trimmed_structure_callback=trimmed_structure_callback,
      strict=strict,
  )
  return result.full_structure if strict else result


def _tree_trim(
    path: tuple[str | int, ...],
    template: PyTreeOf[Any],
    structure: PyTreeOf[T],
    trimmed_structure_callback: TrimmedStructureCallback[T] | None = None,
    strict: bool = True,
) -> PyTreeOf[T]:
  match template:
    # This wants to be `case abc.Mapping()` but http://b/283787842.
    case mapping if isinstance(mapping, abc.Mapping):
      if utils.isinstance_of_namedtuple(structure):
        structure_dict = structure._asdict()  # pytype:disable=attribute-error
      elif isinstance(structure, abc.Mapping):
        structure_dict = structure
      elif structure is None:
        structure_dict = {}
      else:
        raise TypeError(
            f'Type mismatch at key path {path}: template has type'
            f' {type(template)}, but structure has type {type(structure)}.'
        )

      keep_items = []
      drop_items = []
      placeholder_items = []

      if missing := [k for k in template if k not in structure_dict]:
        if strict:
          raise ValueError(
              f'Missing {len(missing)} keys in structure path {path}, '
              f'including: {missing[:10]}'
          )
        else:
          # Fill the result with placeholders
          placeholder_items.extend(
              (k, jax.tree.map(lambda x: ..., template[k])) for k in missing
          )

      for k, n in structure_dict.items():
        (keep_items if k in template else drop_items).append((k, n))

      if trimmed_structure_callback:
        for k, n in drop_items:
          trimmed_structure_callback((*path, k), n)

      keep_dict = {
          k: _tree_trim(
              (*path, k), template[k], v, trimmed_structure_callback, strict
          )
          for k, v in keep_items
      }
      return type(template)((*keep_dict.items(), *placeholder_items))  # pytype:disable=wrong-arg-count
    case named_tuple if utils.isinstance_of_namedtuple(named_tuple):
      if structure is None:
        structure = ()
      if isinstance(structure, abc.Mapping):
        children_dict = _tree_trim(
            path,
            named_tuple._asdict(),
            structure,
            trimmed_structure_callback,
            strict,
        )
        return type(template)(**children_dict)
      elif isinstance(structure, abc.Sequence):
        children_sequence = _tree_trim(
            path,
            tuple(named_tuple),
            structure,
            trimmed_structure_callback,
            strict,
        )
        return type(template)(*children_sequence)
      else:
        raise TypeError(
            f'Type mismatch at key path {path}: template has type'
            f' {type(template)}, but structure has type {type(structure)}.'
        )
    # This wants to be `case abc.Sequence()` but http://b/283787842.
    case sequence if isinstance(sequence, abc.Sequence) and not isinstance(
        sequence, str
    ):
      if structure is None:
        structure = ()
      elif not isinstance(structure, abc.Sequence):
        raise TypeError(
            f'Type mismatch at key path {path}: template has type'
            f' {type(template)}, but structure has type {type(structure)}'
        )
      if len(structure) != len(template):
        raise ValueError(
            f'Length mismatch at key path {path}: template has length'
            f' {len(template)}, but structure has length {len(structure)}'
        )
      elements = (
          _tree_trim((*path, i), t, s, trimmed_structure_callback, strict)
          for i, (t, s) in enumerate(zip(template, structure))
      )
      return type(template)(elements)  # pytype:disable=wrong-arg-count
    case n if n is not None and utils.is_jax_internal_node(n):
      s_children_dict = utils._internal_node_as_dict(structure)  # pylint: disable=protected-access

      t_keys_and_children, t_tree_def = utils.tree_flatten_with_path_one_level(
          template
      )
      t_children_dict = {
          jax.tree_util.keystr(k): v for k, v in t_keys_and_children
      }

      # Note: unlike other cases, this does not treat the children
      # individually. Instead we have effectively cast the structure and
      # the template to mappings and will deal with them in their entirety
      # by reusing the mapping case.
      children_dict = _tree_trim(
          path,
          t_children_dict,
          s_children_dict,
          trimmed_structure_callback,
          strict,
      )
      # Now cast back to the result type.
      children = [
          children_dict[jax.tree_util.keystr(k)] for k, _ in t_keys_and_children
      ]
      return jax.tree_util.tree_unflatten(t_tree_def, children)
    case None:
      # None is special: it's the only type of template tree node that can
      # match both leaves and internal nodes of the structure to be trimmed.
      if utils.is_leaf_node(structure):
        if trimmed_structure_callback:
          trimmed_structure_callback(path, structure)
        return None
      else:
        # Make sure any callback is called appropriately on all elements of
        # `structure`.
        _tree_trim(path, {}, structure, trimmed_structure_callback, strict)
      return None
    case v if utils.is_leaf_node(v):
      if not utils.is_leaf_node_or_none(structure):
        raise TypeError(
            f'Type mismatch at key path {path}: template has type'
            f' {type(template)}, but structure has type {type(structure)}.'
        )
      return structure
    case _:
      raise TypeError(
          f'Unknown type at key path {path}: structure has type'
          f' {type(structure)}.'
      )


def _tree_trim_impl(
    template: PyTreeOf[Any],
    structure: PyTreeOf[T],
    *,
    trimmed_structure_callback: TrimmedStructureCallback[T] | None = None,
    strict: bool = True,
) -> parts_of.PartsOf[PyTreeOf[T]]:
  """Implementation of `tree_trim()` that always returns a `PartsOf`."""
  # To avoid a self-referential recursion, we create a partial that captures
  # the `trimmed_structure_callback` and `strict` arguments instead of doing an
  # implicit closure.
  tree_trim_fn = functools.partial(
      _tree_trim,
      trimmed_structure_callback=trimmed_structure_callback,
      strict=strict,
  )
  return parts_of.PartsOf(template, tree_trim_fn((), template, structure))



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines `CompositeHandler`, a helper component for saving and loading."""

from __future__ import annotations

import asyncio
from typing import Any, Awaitable

from absl import logging
from etils import epath
from orbax.checkpoint._src.metadata import checkpoint as checkpoint_metadata
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import registration
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types


StepMetadata = checkpoint_metadata.StepMetadata
CompositeItemMetadata = checkpoint_metadata.CompositeItemMetadata


def _existing_checkpointable_names(directory: epath.Path) -> set[str]:
  return {p.name for p in directory.iterdir() if p.is_dir()}


async def _create_orbax_identifier_file(
    directory: path_types.PathAwaitingCreation, primary_host: int | None
):
  """Creates a file called `orbax.checkpoint` for easy identification."""
  directory = await directory.await_creation()
  if multihost.is_primary_host(primary_host):
    await asyncio.to_thread(
        (directory / 'orbax.checkpoint').touch, exist_ok=False
    )


class CompositeHandler:
  """CompositeHandler.

  This class is a helper component for `save_checkpointables` and
  `load_checkpointables`. It performs a few core functions:
    - Resolves handlers for saving and loading.
    - Saves and loads checkpointables to/from individual subdirectories by
    delegating to the resolved handlers.

  TODO(b/396190818): This code is only minimally used in the main save/load
  path. It should eventually be incorporated there when the dependency on V0
  implementations is removed.
  """

  def __init__(
      self,
      handler_registry: registration.CheckpointableHandlerRegistry,
  ):
    self._context = context_lib.get_context()
    # Create a copy to prevent mutation by the caller.
    self._handler_registry = registration.local_registry(
        handler_registry, include_global_registry=False
    )
    logging.vlog(
        1,
        'Initialized CompositeHandler with registry: %s.',
        self._handler_registry,
    )

    self._metadata_store = checkpoint_metadata.metadata_store(enable_write=True)

  @property
  def handler_registry(
      self,
  ) -> registration.CheckpointableHandlerRegistry:
    return self._handler_registry

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointables: dict[str, Any],
  ) -> Awaitable[None]:
    """Saves multiple checkpointables to individual subdirectories.

    The subdirectories are expected to already exist.

    Args:
      directory: The directory to save the checkpointables to. The
        checkpointables subdirectories exist under this directory.
      checkpointables: A mapping from checkpointable name to checkpointable.

    Returns:
      An awaitable that represents a background save operation.
    """
    context = context_lib.get_context()
    handlers_for_save = self.get_handlers_for_save(checkpointables)
    save_ops = []
    for checkpointable_name, checkpointable in checkpointables.items():
      save_ops.append(
          handlers_for_save[checkpointable_name].save(
              directory / checkpointable_name, checkpointable
          )
      )
    save_awaitables = await asyncio.gather(*save_ops)

    async def _run_background():
      await asyncio.gather(
          *save_awaitables,
          _create_orbax_identifier_file(
              directory, context.multiprocessing_options.primary_host
          ),
      )

    return _run_background()

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> Awaitable[dict[str, Any]]:
    """Loads multiple checkpointables from individual subdirectories.

    Args:
      directory: The directory to load the checkpointables from. The
        checkpointables subdirectories exist under this directory.
      abstract_checkpointables: A mapping from checkpointable name to abstract
        checkpointable. Only the checkpointables listed will be loaded. If None,
        all checkpointables in the directory will be loaded.

    Returns:
      An awaitable that represents a background load operation.

    Raises:
      KeyError: If any of the specified checkpointable names are not found in
      the checkpoint.
    """
    abstract_checkpointables = abstract_checkpointables or {}
    handlers_for_load = self.get_handlers_for_load(
        directory, abstract_checkpointables
    )
    existing_checkpointable_names = _existing_checkpointable_names(directory)
    if not abstract_checkpointables:
      abstract_checkpointables = {
          name: None
          for name in handlers_for_load.keys()
          if name not in format_utils.RESERVED_CHECKPOINTABLE_KEYS
          and name in existing_checkpointable_names
      }
    if any(
        name not in existing_checkpointable_names
        for name in abstract_checkpointables.keys()
    ):
      raise KeyError(
          'Requested checkpointables for loading were not found in the'
          ' checkpoint. Available checkpointables:'
          f' {existing_checkpointable_names}'
      )

    load_ops = []
    for (
        checkpointable_name,
        abstract_checkpointable,
    ) in abstract_checkpointables.items():
      handler = handlers_for_load[checkpointable_name]
      load_ops.append(
          handler.load(
              directory / checkpointable_name,
              abstract_checkpointable,
          )
      )
    load_awaitables = await asyncio.gather(*load_ops)

    async def _run_background() -> dict[str, Any]:
      loaded_checkpointables = []
      # TODO(b/398249409) Cannot use asyncio.gather because asyncio.run
      # is used in underlying implementation.
      for a in load_awaitables:
        loaded = await a
        loaded_checkpointables.append(loaded)
      return {
          checkpointable_name: loaded
          for checkpointable_name, loaded in zip(
              abstract_checkpointables.keys(),
              loaded_checkpointables,
          )
      }

    return _run_background()

  def get_handlers_for_save(
      self, checkpointables: dict[str, Any]
  ) -> dict[str, handler_types.CheckpointableHandler]:
    """Returns a mapping from checkpointable name to handler."""
    return {
        checkpointable_name: registration.resolve_handler_for_save(
            self._handler_registry, checkpointable, name=checkpointable_name
        )
        for checkpointable_name, checkpointable in checkpointables.items()
    }

  def get_handlers_for_load(
      self, directory: path_types.Path, abstract_checkpointables: dict[str, Any]
  ) -> dict[str, handler_types.CheckpointableHandler]:
    """Returns a mapping from checkpointable name to handler."""
    existing_checkpointable_names_to_handler_typestrs = (
        self._get_saved_handler_typestrs(directory)
    )
    abstract_checkpointables = abstract_checkpointables or {
        name: None for name in existing_checkpointable_names_to_handler_typestrs
    }

    loadable_checkpointable_names_to_handlers = {}
    for name, abstract_checkpointable in abstract_checkpointables.items():
      if name not in existing_checkpointable_names_to_handler_typestrs:
        raise KeyError(
            f'Checkpointable "{name}" was not found in the checkpoint.'
            ' Available names:'
            f' {existing_checkpointable_names_to_handler_typestrs.keys()}'
        )
      handler_typestr = existing_checkpointable_names_to_handler_typestrs[name]
      handler = registration.resolve_handler_for_load(
          self._handler_registry,
          abstract_checkpointable,
          name=name,
          handler_typestr=handler_typestr,
      )
      loadable_checkpointable_names_to_handlers[name] = handler
    return loadable_checkpointable_names_to_handlers

  def _get_saved_handler_typestrs(
      self, directory: path_types.Path
  ) -> dict[str, str]:
    """Reads from the checkpoint metadata to get saved handler typestrs."""
    step_metadata_file_path = checkpoint_metadata.step_metadata_file_path(
        directory
    )
    if step_metadata_file_path.exists():
      serialized_metadata = self._metadata_store.read(step_metadata_file_path)
      saved_metadata = step_metadata_serialization.deserialize(
          serialized_metadata or {}
      )
      if isinstance(saved_metadata.item_handlers, dict):
        return saved_metadata.item_handlers  # found step level metadata.
      raise ValueError(
          'Expected a valid path containing checkpointable subdirectories, but'
          ' given path contains subdirectories:'
          f' {format_utils.subdirs(directory)}... Given path is {directory}.'
          ' _CHECKPOINT_METADATA file under given path has'
          f' `item_handlers`={saved_metadata.item_handlers}, whose keys should'
          ' match the checkpointable subdirectory names. If you intended to'
          ' load a pytree checkpoint from the given path, then please consider'
          ' using `loading.load_pytree(..., checkpointable_name=None)` instead.'
      )

    logging.warning(
        'Given dir does not contain checkpoint metadata file: %s. Trying to get'
        ' saved handlers from checkpoint metadata in each of the checkpointable'
        ' subdirectory.',
        directory,
    )

    saved_handler_typestrs: dict[str, str] = {}
    for checkpointable_path in directory.iterdir():
      serialized_metadata = self._metadata_store.read(
          checkpoint_metadata.step_metadata_file_path(checkpointable_path)
      )
      if serialized_metadata is None:
        continue
      saved_metadata = step_metadata_serialization.deserialize(
          serialized_metadata
      )
      if isinstance(saved_metadata.item_handlers, dict):
        raise ValueError(
            'Expected a valid path containing checkpointable subdirectories,'
            ' but given path contains subdirectories:'
            f' {format_utils.subdirs(directory)}... Given path is {directory}.'
            ' _CHECKPOINT_METADATA file under a subdir of given path has'
            f' `item_handlers`={saved_metadata.item_handlers}, whose keys'
            ' should match the checkpointable subdirectory names. Did you mean'
            ' to provide the following subdirectory path instead:'
            f' {checkpointable_path}?'
        )
      item_handlers = saved_metadata.item_handlers
      if item_handlers is not None:
        checkpointable_name = checkpointable_path.name
        saved_handler_typestrs[checkpointable_name] = item_handlers
    return saved_handler_typestrs



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at root level."""
from __future__ import annotations

import dataclasses
from typing import Any

from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
RootMetadata = checkpoint.RootMetadata


@dataclasses.dataclass
class InternalRootMetadata:
  """Internal representation of `RootMetadata`.

  See documentation on `InternalCheckpointMetadata` for more design context.
  """

  custom_metadata: dict[str, Any] | None = dataclasses.field(
      default_factory=dict,
      metadata={'processor': utils.validate_and_process_custom_metadata},
  )

  def serialize(self) -> SerializedMetadata:
    """Serializes `self` to a JSON dictionary and performs validation."""
    return {
        field.name: field.metadata['processor'](getattr(self, field.name))
        for field in dataclasses.fields(self)
    }

  @classmethod
  def deserialize(
      cls, metadata_dict: SerializedMetadata
  ) -> InternalRootMetadata:
    """Deserializes `metadata_dict` to `InternalRootMetadata`."""
    assert isinstance(metadata_dict, dict)
    fields = dataclasses.fields(InternalRootMetadata)
    field_names = {field.name for field in fields}

    field_processor_args = {
        field_name: (metadata_dict.get(field_name, None),)
        for field_name in field_names
    }

    validated_metadata_dict = {
        field.name: field.metadata['processor'](
            *field_processor_args[field.name]
        )
        for field in fields
    }

    for k in metadata_dict:
      if k not in validated_metadata_dict:
        validated_metadata_dict['custom_metadata'][k] = (
            utils.process_unknown_key(k, metadata_dict)
        )
    return InternalRootMetadata(**validated_metadata_dict)

  def to_root_metadata(self) -> RootMetadata:
    return RootMetadata(custom_metadata=self.custom_metadata)

  @classmethod
  def from_root_metadata(
      cls, root_metadata: RootMetadata
  ) -> InternalRootMetadata:
    return InternalRootMetadata(custom_metadata=root_metadata.custom_metadata)


def serialize(metadata: RootMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""
  return InternalRootMetadata.from_root_metadata(metadata).serialize()


def deserialize(metadata_dict: SerializedMetadata) -> RootMetadata:
  """Deserializes `metadata_dict` to `RootMetadata`."""
  return InternalRootMetadata.deserialize(metadata_dict).to_root_metadata()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""StringLeafHandler that implements the types.LeafHandler Protocol.

The primary purpose of this handler is to provide serialization and
deserialization for strings.
"""

import asyncio
from typing import Awaitable, Sequence, Type

from absl import logging
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.serialization import type_handlers as type_handlers_v0
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.serialization import types


AbstractString = Type[str]
StringSerializationParam = types.SerializationParam[str]
StringDeserializationParam = types.DeserializationParam[AbstractString]


def _create_v0_saving_paraminfo(
    param: StringSerializationParam,
    context: context_lib.Context,
    serialization_context: types.SerializationContext,
) -> type_handlers_v0.ParamInfo:
  """Creates a V0 ParamInfo from V1 params and contexts for saving."""

  saving_options = context.array_options.saving

  return type_handlers_v0.ParamInfo(
      name=param.name,
      path=serialization_context.parent_dir.path / param.name,
      parent_dir=serialization_context.parent_dir.path,
      byte_limiter=serialization_context.byte_limiter,
      is_ocdbt_checkpoint=saving_options.use_ocdbt,
      use_zarr3=saving_options.use_zarr3,
      ocdbt_target_data_file_size=saving_options.ocdbt_target_data_file_size,
      ts_context=serialization_context.ts_context,
      value_typestr="string",
  )


def _create_v0_restore_paraminfo(
    param: types.DeserializationParam[None | AbstractString],
    context: context_lib.Context,
    deserialization_context: types.DeserializationContext,
) -> type_handlers_v0.ParamInfo:
  """Creates a V0 ParamInfo from V1 params and contexts for loading."""

  loading_options = context.array_options.Loading

  return type_handlers_v0.ParamInfo(
      name=param.name,
      path=deserialization_context.parent_dir / param.name,
      parent_dir=deserialization_context.parent_dir,
      skip_deserialize=False,
      byte_limiter=deserialization_context.byte_limiter,
      is_ocdbt_checkpoint=deserialization_context.ocdbt_checkpoint,
      ts_context=deserialization_context.ts_context,
      raise_array_data_missing_error=loading_options.raise_array_data_missing_error,
      use_zarr3=deserialization_context.zarr3_checkpoint,
  )


async def _async_futures(commit_futures: Sequence[future.Future]):
  await asyncio.gather(*[asyncio.to_thread(f.result) for f in commit_futures])


class StringLeafHandler(types.LeafHandler[str, AbstractString]):
  """StringLeafHandler that implements the types.LeafHandler Protocol."""

  def __init__(
      self,
      *,
      context: context_lib.Context | None = None,
  ):
    """Initializes the StringLeafHandler.

    This handler underneath uses the V0 StringHandler.

    Args:
      context: Context that will be used for this leaf handler.
    """
    self._context = context_lib.get_context(context)
    self._handler_impl = type_handlers_v0.StringHandler()

    logging.vlog(1, "StringLeafHandler created.")

  async def serialize(
      self,
      params: Sequence[StringSerializationParam],
      serialization_context: types.SerializationContext,
  ) -> Awaitable[None]:
    """Serializes scalar values as a checkpointable to a storage location.

    Args:
      params: a sequence of StringSerializationParam per leaf.
      serialization_context: SerializationContext for the scalar leaf handler.

    Returns:
      Sequence of commit futures which can be awaited to complete the save
      operation.
    """
    values = [p.value for p in params]
    paraminfos = [
        _create_v0_saving_paraminfo(p, self._context, serialization_context)
        for p in params
    ]

    # `args` is not used by StringHandler.serialize, so it's not passed in.
    commit_futures = await self._handler_impl.serialize(values, paraminfos)
    if not commit_futures:
      raise ValueError("No commit futures returned by StringHandler.serialize.")

    return _async_futures(commit_futures)

  async def deserialize(
      self,
      params: Sequence[types.DeserializationParam[AbstractString]],
      deserialization_context: types.DeserializationContext,
  ) -> Awaitable[Sequence[str]]:
    """Returns sequence of String values from a stored checkpointable location.

    Args:
      params: sequence of StringDeserializationParam per leaf.
      deserialization_context: StringDeserializationContext for the leaf
        handler.

    Returns:
      The deserialized sequence of scalar values as leaves.
    """

    # validate all parameters
    paraminfos = [
        _create_v0_restore_paraminfo(p, self._context, deserialization_context)
        for p in params
    ]

    async def _background_deserialize() -> Sequence[str]:
      # This is needed because StringHandler.deserialize could return None
      # values.  However, it should be very rare.  This is to make sure
      # everything is string.
      return [p or "" for p in await self._handler_impl.deserialize(paraminfos)]

    return asyncio.create_task(_background_deserialize())

  async def metadata(
      self,
      params: Sequence[types.DeserializationParam[None]],
      deserialization_context: types.DeserializationContext,
  ) -> Sequence[AbstractString]:
    """Returns a squence of str from a stored checkpointable location.

    Args:
      params: sequence of StringDeserializationParam per scalar value leaf.
      deserialization_context: DeserializationContext for the scalar leaf
        handler.

    Returns:
      Sequence of StringMetadata for each provided DeserializationParam.
    """
    paraminfos = [
        _create_v0_restore_paraminfo(p, self._context, deserialization_context)
        for p in params
    ]

    async def _get_metadata() -> Sequence[AbstractString]:
      v0_metadatas = await self._handler_impl.metadata(paraminfos)
      return [str] * len(v0_metadatas)

    return await _get_metadata()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public symbols for type_handlers module."""

# pylint: disable=g-importing-member, unused-import, g-bad-import-order

from orbax.checkpoint._src.serialization.type_handlers import ArrayHandler
from orbax.checkpoint._src.serialization.type_handlers import ArrayRestoreArgs
from orbax.checkpoint._src.serialization.type_handlers import NumpyHandler
from orbax.checkpoint._src.serialization.type_handlers import ParamInfo
from orbax.checkpoint._src.serialization.type_handlers import RestoreArgs
from orbax.checkpoint._src.serialization.type_handlers import SaveArgs
from orbax.checkpoint._src.serialization.type_handlers import ScalarHandler
from orbax.checkpoint._src.serialization.type_handlers import SingleReplicaArrayHandler
from orbax.checkpoint._src.serialization.type_handlers import SingleReplicaArrayRestoreArgs
from orbax.checkpoint._src.serialization.type_handlers import StringHandler
from orbax.checkpoint._src.serialization.type_handlers import TypeHandler

# TypeHandler Registry
from orbax.checkpoint._src.serialization.type_handlers import TypeHandlerRegistry
from orbax.checkpoint._src.serialization.type_handlers import create_type_handler_registry
from orbax.checkpoint._src.serialization.type_handlers import get_type_handler
from orbax.checkpoint._src.serialization.type_handlers import has_type_handler
from orbax.checkpoint._src.serialization.type_handlers import register_standard_handlers_with_options
from orbax.checkpoint._src.serialization.type_handlers import register_type_handler
from orbax.checkpoint._src.serialization.type_handlers import supported_types

from orbax.checkpoint._src.serialization.type_handlers import is_ocdbt_checkpoint

# BEGIN OF DEPRECATED FUNCTIONS
# DON'T USE UNLESS YOU KNOW WHAT YOU'RE DOING
from orbax.checkpoint._src.serialization.type_handlers import _assert_parameter_files_exist
from orbax.checkpoint._src.serialization.type_handlers import _get_json_tspec
from orbax.checkpoint._src.serialization.type_handlers import check_input_arguments
from orbax.checkpoint._src.serialization.type_handlers import get_cast_tspec_deserialize
from orbax.checkpoint._src.serialization.type_handlers import get_cast_tspec_serialize
from orbax.checkpoint._src.serialization.type_handlers import get_json_tspec_read
from orbax.checkpoint._src.serialization.type_handlers import get_json_tspec_write
from orbax.checkpoint._src.serialization.type_handlers import get_ts_context
from orbax.checkpoint._src.serialization.type_handlers import is_supported_empty_value
from orbax.checkpoint._src.serialization.type_handlers import is_supported_type
from orbax.checkpoint._src.serialization.type_handlers import merge_ocdbt_per_process_files
# END OF DEPRECATED FUNCTIONS



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides definitions for AggregateHandler and implementations."""

import abc
from typing import Any, Optional

from etils import epath
from orbax.checkpoint import future as orbax_future
from orbax.checkpoint import msgpack_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src.metadata import tree as tree_metadata

PyTree = Any


class AggregateHandler(abc.ABC):
  """Interface for reading and writing a PyTree using a specific format."""

  @abc.abstractmethod
  async def serialize(
      self, path: epath.Path, item: PyTree
  ) -> orbax_future.Future:
    """Serializes and writes `item` to a given `path`.

    The function is compatible with a multihost setting, but does not include
    extra logic to ensure atomicity.

    Args:
      path: the folder to which the item should be written.
      item: a PyTree.
    """
    pass

  @abc.abstractmethod
  def deserialize(self, path: epath.Path) -> PyTree:
    """Reads and deserializes a PyTree from the given directory."""
    pass

  @abc.abstractmethod
  def close(self):
    """Closes the handler."""
    pass


class MsgpackHandler(AggregateHandler):
  """An implementation of AggregateHandler that uses msgpack to store the tree."""

  def __init__(
      self,
      primary_host: Optional[int] = 0,
      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
          tree_metadata.PYTREE_METADATA_OPTIONS
      ),
  ):
    self._primary_host = primary_host
    self._pytree_metadata_options = pytree_metadata_options

  async def serialize(
      self, path: epath.Path, item: PyTree
  ) -> orbax_future.Future:
    """See superclass documentation."""

    async def _serialize_fn(x):
      if utils.is_primary_host(self._primary_host):
        if self._pytree_metadata_options.support_rich_types:
          raise NotImplementedError(
              'Orbax does not support rich typed metadata in legacy msgpack'
              ' checkpoint format. Please set'
              ' PyTreeMetadataOptions.support_rich_types to False.'
          )
        serializable_dict = tree_metadata.serialize_tree(
            x, self._pytree_metadata_options
        )
        msgpack = msgpack_utils.msgpack_serialize(serializable_dict)
        # Explicit "copy" phase is not needed because msgpack only contains
        # basic types and numpy arrays.
        return path.write_bytes(msgpack)

    return orbax_future.CommitFutureAwaitingContractedSignals(
        _serialize_fn(item)
    )

  def deserialize(self, path: epath.Path) -> PyTree:
    """See superclass documentation."""
    if path.exists():
      msgpack = path.read_bytes()
      return msgpack_utils.msgpack_restore(msgpack)
    else:
      raise FileNotFoundError(f'Checkpoint does not exist at {path}.')

  def close(self):
    """See superclass documentation."""
    pass



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler that combines other handlers.

Usage example::

  import orbax.checkpoint as ocp

  # A PyTree of jax.Arrays
  my_state = ...
  # A dictionary to be serialized as JSON
  json_dict = ...

  checkpointer = ocp.Checkpointer(ocp.CompositeCheckpointHandler())
  checkpointer.save(
      path,
      args=ocp.args.Composite(
          state=ocp.args.StandardSave(my_state),
          metadata=ocp.args.JsonSave(json_dict)
      )
  )

  restored = checkpointer.restore(
      path,
      args=ocp.args.Composite(
          state=ocp.args.StandardRestore(),
          metadata=ocp.args.JsonRestore()
      )
  )
  my_state = restored.state
  json_dict = restored.metadata
"""

from __future__ import annotations

import asyncio
import concurrent.futures
import dataclasses
import threading
import time
from typing import Any, Dict, List, Mapping, MutableSet, Optional, Tuple, Type

from absl import logging
from etils import epath
import jax
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src import composite
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.handlers import async_checkpoint_handler
from orbax.checkpoint._src.handlers import checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration
from orbax.checkpoint._src.handlers import handler_type_registry
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import atomicity_defaults
from orbax.checkpoint._src.path import atomicity_types


CheckpointArgs = checkpoint_args.CheckpointArgs
Future = future.Future
CheckpointArgs = checkpoint_args.CheckpointArgs
CheckpointHandler = checkpoint_handler.CheckpointHandler
StepMetadata = checkpoint.StepMetadata
CompositeItemMetadata = checkpoint.CompositeItemMetadata
AsyncCheckpointHandler = async_checkpoint_handler.AsyncCheckpointHandler
register_with_handler = checkpoint_args.register_with_handler
Composite = composite.Composite
HandlerAwaitableSignal = synchronization.HandlerAwaitableSignal


_CONCURRENT_WORKERS = 3

DESCRIPTOR_ITEM_NAME = 'descriptor'
RESERVED_ITEM_NAMES = []
_DIRECTORY_CREATION_SIGNALS = [HandlerAwaitableSignal.ITEM_DIRECTORY_CREATION]


# TODO(b/295899152) Clean up when users are all registering `CheckpointArgs`.
class _LegacyCheckpointHandlerWrapper(checkpoint_handler.CheckpointHandler):
  """Wrapper for `CheckpointHandler`s without registered `CheckpointArgs`."""

  def __init__(self, handler: checkpoint_handler.CheckpointHandler):
    self._handler = handler

  def save(self, directory: epath.Path, args: _WrapperArgs):
    return self._handler.save(directory, *args.args, **args.kwargs)

  def restore(self, directory: epath.Path, args: _WrapperArgs):
    return self._handler.restore(directory, *args.args, **args.kwargs)

  def metadata(self, directory: epath.Path) -> Optional[Any]:
    return self._handler.metadata(directory)

  def structure(self, directory: epath.Path) -> Optional[Any]:
    if hasattr(self._handler, 'structure'):
      return self._handler.structure(directory)
    raise AttributeError(
        f'CheckpointHandler of type: {type(self._handler)} has no method'
        ' `structure`.'
    )

  def finalize(self, directory: epath.Path):
    return self._handler.finalize(directory)

  def close(self):
    return self._handler.close()


@register_with_handler(
    _LegacyCheckpointHandlerWrapper, for_save=True, for_restore=True
)
class _WrapperArgs(CheckpointArgs):

  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs


# TODO(b/295899152) Clean up when users are all registering `CheckpointArgs`.
class _AsyncLegacyCheckpointHandlerWrapper(
    async_checkpoint_handler.AsyncCheckpointHandler
):
  """Wrapper for `CheckpointHandler`s without registered `CheckpointArgs`."""

  def __init__(self, handler: async_checkpoint_handler.AsyncCheckpointHandler):
    self._handler = handler

  def save(self, directory: epath.Path, args: _AsyncWrapperArgs):
    return self._handler.save(directory, *args.args, **args.kwargs)

  async def async_save(self, directory: epath.Path, args: _AsyncWrapperArgs):
    return await self._handler.async_save(directory, *args.args, **args.kwargs)

  def restore(self, directory: epath.Path, args: _AsyncWrapperArgs):
    return self._handler.restore(directory, *args.args, **args.kwargs)

  def metadata(self, directory: epath.Path) -> Optional[Any]:
    return self._handler.metadata(directory)

  def structure(self, directory: epath.Path) -> Optional[Any]:
    if hasattr(self._handler, 'structure'):
      return self._handler.structure(directory)
    raise AttributeError(
        f'CheckpointHandler of type: {type(self._handler)} has no method'
        ' `structure`.'
    )

  def finalize(self, directory: epath.Path):
    return self._handler.finalize(directory)

  def close(self):
    return self._handler.close()


@register_with_handler(
    _AsyncLegacyCheckpointHandlerWrapper, for_save=True, for_restore=True
)
class _AsyncWrapperArgs(CheckpointArgs):

  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs


def get_legacy_handler_wrapper(
    handler: checkpoint_handler.CheckpointHandler,
) -> checkpoint_handler.CheckpointHandler:
  if isinstance(handler, async_checkpoint_handler.AsyncCheckpointHandler):
    return _AsyncLegacyCheckpointHandlerWrapper(handler)
  return _LegacyCheckpointHandlerWrapper(handler)


def _maybe_raise_reserved_item_error(item_name: str):
  if item_name in RESERVED_ITEM_NAMES:
    raise ValueError(f'Cannot specify reserved item name: "{item_name}".')


@dataclasses.dataclass
class CompositeOptions:
  multiprocessing_options: options_lib.MultiprocessingOptions = (
      dataclasses.field(default_factory=options_lib.MultiprocessingOptions)
  )
  temporary_path_class: Optional[Type[atomicity_types.TemporaryPath]] = None
  file_options: Optional[options_lib.FileOptions] = None
  async_options: Optional[options_lib.AsyncOptions] = None


def _get_unique_registered_items_and_handlers(
    registry: handler_registration.CheckpointHandlerRegistry,
) -> List[Tuple[str, CheckpointHandler]]:
  """Returns unique items and handlers from the registry.

  Args:
    registry: The registry to get the items and handlers from.

  Returns:
    A list of unique `(item name, handler)` tuples.
  """
  item_and_handers = []
  for (
      item,
      _,
  ), handler in registry.get_all_entries().items():
    if item is not None and (item, handler) not in item_and_handers:
      item_and_handers.append((item, handler))
  return item_and_handers


def _add_to_handler_registry_from_global_registry(
    registry: handler_registration.CheckpointHandlerRegistry,
    registered_handler_for_args: CheckpointHandler,
    item_name: str,
) -> None:
  """Adds items and handlers to a registry from the global registry."""

  save_args, restore_args = checkpoint_args.get_registered_args_cls(
      registered_handler_for_args
  )

  logging.info(
      'Deferred registration for item: "%s". Adding handler `%s` for item "%s"'
      ' and save args `%s` and restore args `%s` to `_handler_registry`.',
      item_name,
      registered_handler_for_args,
      item_name,
      save_args,
      restore_args,
  )
  registry.add(item_name, save_args, registered_handler_for_args)

  # If the restore args are different from the save args, add them to the
  # registry as well.
  if restore_args != save_args:
    registry.add(item_name, restore_args, registered_handler_for_args)


class CompositeCheckpointHandler(AsyncCheckpointHandler):
  """CheckpointHandler for saving multiple items.

  As with all :py:class:`.CheckpointHandler` implementations, use only in
  conjunction with an instance of :py:class:`.AbstractCheckpointer`.

  The :py:class:`.CompositeCheckpointHandler` allows dealing with multiple items
  of different types or logical distinctness, such as training state (PyTree),
  dataset iterator, JSON metadata, or anything else. To specify the handler
  and args for each checkpointable item, use the `handler_registry` option at
  initialization (see :py:class:`.CheckpointHandlerRegistry`).

  If a handler registry is provided, the :py:class:`.CompositeCheckpointHandler`
  will use this registry to determine the :py:class:`.CheckpointHandler` for
  each checkpointable item. If no registry is provided, an empty registry will
  be constructed by default internally. Additional items not registered in the
  handler registry can be specified when calling `save` or `restore`. The
  handler will be determined from the provided :py:class:`.CheckpointArgs`
  during the first call to `save` or `restore`. Subsequent calls to `save` or
  `restore` will use the same handler and will require the same
  :py:class:`.CheckpointArgs` to be provided.

  `item_names` and `items_and_handlers` are deprecated arguments for specifying
  the items and handlers. Please use `handler_registry` instead.

  Usage::

    # The simplest use-case, with no handler registry provided on construction.
    checkpointer = ocp.Checkpointer(
        ocp.CompositeCheckpointHandler()
    )

    checkpointer.save(directory,
        ocp.args.Composite(
            # The handler will be determined from the global registry using the
            # provided args.
            state=ocp.args.StandardSave(pytree),
        )
    )

    restored: ocp.args.Composite = checkpoint_handler.restore(directory,
        ocp.args.Composite(
            state=ocp.args.StandardRestore(abstract_pytree),
        )
    )
    restored.state ...

    # The "state" item's args was determined on the first call to `save`.
    # Trying to save "state" with a different set of args will raise an error.
    checkpointer.save(directory,
      ocp.args.Composite(
        # Will raise `ValueError: Item "state" and args "JsonSave [...]" does
        # not match with any registered handler! ...`.
        state=ocp.args.JsonSave(pytree),
        )
    )

    # You can also register specific items and handlers using a handler
    # registry.
    handler_registry = (
        ocp.handler_registration.DefaultCheckpointHandlerRegistry()
    )
    # Some subclass of `CheckpointHandler` that handles the "state" item.
    handler = state_handler()
    # Multiple save args and handlers can be registered for the same item.
    handler_registry.add(
        'state,
        StateSaveArgs,
        handler,
    )
    handler_registry.add(
        'state',
        StateRestoreArgs,
        handler,
    )

    # The `CompositeCheckpointHandler` will use the handlers registered in the
    # registry when it encounters the item name `state`.
    checkpointer_with_registry = ocp.Checkpointer(
        ocp.CompositeCheckpointHandler(handler_registry=handler_registry)
    )

    checkpointer_with_registry.save(directory,
        ocp.args.Composite(
            # The handler knows that the "state" item should use the
            # `state_handler` specified in the registry. When an item has been
            # added to the handler registry, only save args that are registered
            # for that item are allowed.
            state=StateSaveArgs(pytree),
            # You can also provide other arguments that are not registered in
            # the handler registry as long as a handler has been globally
            # registered for the args.
            other_param=ocp.args.StandardSave(other_pytree),
        )
    )

    restored: ocp.args.Composite = checkpointer_with_registry.restore(directory,
        ocp.args.Composite(
            state=StateRestoreArgs(abstract_pytree),
            other_param=ocp.args.StandardRestore(abstract_other_pytree),
        )
    )
    restored.state ...
    restored.other_param...

    # Skip restoring `other_params` (you can save a subset of items too, in a
    # similar fashion).
    restored_state_only: ocp.args.Composite = (
      checkpointer_with_registry.restore(directory,
        ocp.args.Composite(
            state=StateRestoreArgs(abstract_pytree),
        )
      )
    )
    restored.state ...
    restored.other_param ... # None.
  """

  _current_temporary_paths: Dict[str, atomicity_types.TemporaryPath] = {}
  # Items that do not have a registered handler. This is set to `None` if the
  # user provided a `handler_registry` at initialization. If the user provided
  # `item_names` and `items_and_handlers` at initialization, this is set to a
  # set of the item names that have a `None` handler and will need to be
  # registered at the first call to `_get_or_set_handler`.
  _item_names_without_registered_handlers: Optional[MutableSet[str]] = None

  def __init__(
      self,
      *item_names: str,
      composite_options: CompositeOptions = CompositeOptions(),
      handler_registry: Optional[
          handler_registration.CheckpointHandlerRegistry
      ] = None,
      **items_and_handlers: CheckpointHandler,
  ):
    """Constructor.

    If you are `item_names` and/or `items_and_handlers`, all items must be
    provided up-front, at initialization. If you are using a `handler_registry`,
    you can register items at any time, even after the first call to `save` or
    `restore`.

    Args:
      *item_names: A list of string item names that this handler will manage.
        `item_names` is deprecated. Please use `handler_registry` instead.
      composite_options: Options.
      handler_registry: A :py:class:`CheckpointHandlerRegistry` instance. If
        provided, the :py:class:`CompositeCheckpointHandler` will use this
        registry to determine the :py:class:`CheckpointHandler` for each item.
        This option is mutually exclusive with `items_and_handlers` and
        `item_names`.
      **items_and_handlers: A mapping of item name to `CheckpointHandler`
        instance, which will be used as the handler for objects of the
        corresponding name. `items_and_handlers` is deprecated. Please use
        `handler_registry` instead.
    """
    if handler_registry is not None and items_and_handlers:
      raise ValueError(
          'Both `handler_registry` and `items_and_handlers` were provided. '
          'Please specify only one of the two.'
      )
    if handler_registry is not None and item_names:
      raise ValueError(
          'Both `handler_registry` and `item_names` were provided. '
          'Please specify only one of the two.'
      )
    if items_and_handlers or item_names:
      # This whole code branch will be removed once `item_and_handlers` and
      # `item_names` have been completely deprecated.
      self._item_names_without_registered_handlers: MutableSet[str] = set()
      self._handler_registry = (
          handler_registration.DefaultCheckpointHandlerRegistry()
      )

      if item_names:
        for item in item_names:
          _maybe_raise_reserved_item_error(item)
          if item not in items_and_handlers:
            # If an item is in `item_names`, but not in `items_and_handlers`,
            # there is no way to know what the corresponding handler and args
            # are. Instead, defer registration until the first call to
            # `_get_or_set_handler`.
            self._item_names_without_registered_handlers.add(item)

      # Register all the items and handlers in the handler registry.
      if items_and_handlers:
        for item, handler in items_and_handlers.items():
          _maybe_raise_reserved_item_error(item)
          if handler is not None and checkpoint_args.has_registered_args(
              handler
          ):
            _add_to_handler_registry_from_global_registry(
                self._handler_registry,
                handler,
                item,
            )
          elif handler is not None:
            logging.warning(
                'No registered CheckpointArgs found for handler type: %s',
                type(handler),
            )
            legacy_wrapped_handler = get_legacy_handler_wrapper(handler)
            self._handler_registry.add(
                item,
                _WrapperArgs,
                legacy_wrapped_handler,
            )
          else:
            # If the handler is `None`, the handler will be determined from the
            # global checkpoint args registry during the first call to
            # `_get_or_set_handler`.
            self._item_names_without_registered_handlers.add(item)

      self._items_with_deferred_registration = None
    elif handler_registry is not None:
      # Create a new handler registry to prevent the input from being mutated.
      self._handler_registry = (
          handler_registration.DefaultCheckpointHandlerRegistry(
              handler_registry
          )
      )
      self._item_names_without_registered_handlers = None
      self._items_with_deferred_registration = set()

      # Prevent the user from lazily registering items that are already
      # registered in the provided handler registry.
      for item, _ in _get_unique_registered_items_and_handlers(
          self._handler_registry
      ):
        self._items_with_deferred_registration.add(item)
    else:
      self._handler_registry = (
          handler_registration.DefaultCheckpointHandlerRegistry()
      )
      self._item_names_without_registered_handlers = None
      self._items_with_deferred_registration = set()

    self._primary_host = composite_options.multiprocessing_options.primary_host
    self._active_processes = (
        composite_options.multiprocessing_options.active_processes
    )
    self._barrier_sync_key_prefix = (
        composite_options.multiprocessing_options.barrier_sync_key_prefix
    )
    self._multiprocessing_options = composite_options.multiprocessing_options

    self._temporary_path_class = composite_options.temporary_path_class
    self._file_options = composite_options.file_options
    self._async_options = (
        composite_options.async_options or options_lib.AsyncOptions()
    )
    logging.info(
        'Initialized registry %s.',
        self._handler_registry,
    )

    self._metadata_store = checkpoint.metadata_store(enable_write=False)

  def _get_all_registered_and_unregistered_items_and_handlers(
      self,
  ) -> list[tuple[str, Optional[CheckpointHandler]]]:
    """Gets all items and corresponding handlers.

    Can be rmeoved once `item_names` and `items_and_handlers` have been
    completely deprecated.

    Returns:
      A tuple containing all uniuqe items and handlers registered in the handler
      registry. If `item_names` or `items_and_handlers` has been specified, the
      tuple may also contain item without a corresponding checkpoint handler.
    """
    items_and_handlers = _get_unique_registered_items_and_handlers(
        self._handler_registry
    )
    if self._item_names_without_registered_handlers is not None:
      for item in self._item_names_without_registered_handlers:
        items_and_handlers.append((item, None))
    return items_and_handlers

  def _get_or_set_handler(
      self,
      item_name: str,
      args: Optional[CheckpointArgs],
  ) -> CheckpointHandler:
    handler_registry = self._handler_registry
    assert handler_registry is not None
    if args is not None:
      # Check if registry contains a handler for the item name or the
      # the general argument type.
      if handler_registry.has(item_name, args):
        return handler_registry.get(item_name, args)
      elif handler_registry.has(None, args):
        handler = handler_registry.get(None, args)
        # There is a default handler for the args, but it is not registered
        # for the specific item. Register the handler for the item.
        handler_registry.add(item_name, args, handler)
        return handler
      else:
        logging.info(
            'No entry found in handler registry for item: %s and args with'
            ' type: %s. Falling back to global handler registry.',
            item_name,
            type(args),
        )
    else:
      # Try to find a handler for the item. If none is found, raise an error.
      items_and_handlers = _get_unique_registered_items_and_handlers(
          handler_registry
      )
      for item, handler in items_and_handlers:
        if item_name == item:
          assert handler is not None
          return handler
      raise ValueError(
          'Provided `None` for `CheckpointArgs`, and the `CheckpointHandler`'
          f' for item "{item_name}" was not configured. If saving, this'
          f' indicates that "{item_name}" was saved with `None` (an instance'
          ' of `CheckpointArgs` was expected). If restoring, providing'
          ' `None` for the item is valid, but only if the `CheckpointHandler'
          ' was already configured. Provide a `CheckpointArgs` subclass so'
          ' that `CompositeCheckpointHandler` will know how to restore the'
          ' item.'
      )

    # Retrieve and default initialize the handler from the global registry. This
    # handler will be added to the handler registry if it is not already
    # registered for the item.
    registered_handler_for_args = (
        checkpoint_args.get_registered_handler_cls(args)
    )()  # pytype: disable=not-instantiable

    # Add items from `item_names` to the handler registry if it is not already
    # registered. This happens on the first call of `_get_or_set_handler` for
    # each item. `self._item_names_without_registered_handlers` is only set if
    # the user provided `item_names` or `items_and_handlers` at initialization.
    item_names_without_registered_handlers = (
        self._item_names_without_registered_handlers
    )
    if item_names_without_registered_handlers is not None:
      if item_name in item_names_without_registered_handlers:
        item_names_without_registered_handlers.remove(item_name)
        _add_to_handler_registry_from_global_registry(
            handler_registry,
            registered_handler_for_args,
            item_name,
        )
        return registered_handler_for_args
    else:
      # This branch is only reached if the user provided a `handler_registry` at
      # initialization. Any item that is not registered in the provided handler
      # registry will be registered in a deferred manner. This allows the user
      # to provide a `handler_registry` at initialization that does not need to
      # contain all the items that will be checkpointed, as long as the
      # checkpoint args for the item have been registered globally with
      # a corresponding handler.
      assert self._items_with_deferred_registration is not None
      if item_name not in self._items_with_deferred_registration:
        # Add the item to the set of items with deferred registration to prevent
        # adding the same item multiple times to the handler registry.
        self._items_with_deferred_registration.add(item_name)
        _add_to_handler_registry_from_global_registry(
            handler_registry,
            registered_handler_for_args,
            item_name,
        )
        return registered_handler_for_args

    raise ValueError(
        f'Item "{item_name}" and args "{args}" does not match with any'
        ' registered handler! Registered handlers:'
        f' {self._handler_registry.get_all_entries().items()}'
    )

  def _get_item_directory(
      self, directory: epath.Path, item_name: str
  ) -> epath.Path:
    return directory / item_name

  def _get_item_temporary_directory(
      self, directory: epath.Path, item_name: str
  ) -> atomicity_types.TemporaryPath:
    temporary_path_class = (
        self._temporary_path_class
        or atomicity_defaults.get_item_default_temporary_path_class(directory)
    )
    tmp_item_dir = temporary_path_class.from_final(
        self._get_item_directory(directory, item_name),
        multiprocessing_options=self._multiprocessing_options,
        file_options=self._file_options,
    )
    return tmp_item_dir

  def _get_item_temporary_paths(
      self, directory: epath.Path, args: CompositeArgs
  ) -> Dict[str, atomicity_types.TemporaryPath]:
    # Sort keys to maintain consistent ordering across processes, otherwise
    # we may hit timeouts if processes wait at different barriers in per-item
    # handlers.
    # TODO(b/295899152): Find a less brittle solution or support
    # async-compatible barrier function within handlers.
    # The main blocker here is that many users with custom CheckpointHandlers
    # use barriers in their implementations, which are usually not actually
    # needed.
    item_names = sorted(args.keys())
    item_temporary_paths = [
        self._get_item_temporary_directory(directory, item_name)
        for item_name in item_names
    ]
    result = {
        item_name: item_directory
        for item_name, item_directory in zip(item_names, item_temporary_paths)
    }

    return result

  async def async_save(
      self, directory: epath.Path, args: CompositeArgs
  ) -> Optional[List[Future]]:
    """Saves multiple items to individual subdirectories."""
    start_time = time.time()
    self._current_temporary_paths = self._get_item_temporary_paths(
        directory, args
    )
    commit_futures = []
    if self._async_options.create_directories_asynchronously:
      commit_futures.append(
          atomicity.create_all_async(
              list(self._current_temporary_paths.values()),
              completion_signals=_DIRECTORY_CREATION_SIGNALS,
              multiprocessing_options=self._multiprocessing_options,
          )
      )
    else:
      future.CommitFutureAwaitingContractedSignals(
          atomicity.create_all(
              list(self._current_temporary_paths.values()),
              multiprocessing_options=self._multiprocessing_options,
          )
      ).result()
    save_ops = []
    time_taken_per_item = {}
    time_taken_all_items = 0.0
    for item_name, item_directory in self._current_temporary_paths.items():
      item_start_time = time.time()
      arg = args[item_name]
      _maybe_raise_reserved_item_error(item_name)
      handler = self._get_or_set_handler(item_name, arg)
      if isinstance(handler, AsyncCheckpointHandler):
        save_ops.append(handler.async_save(item_directory.get(), args=arg))
      else:
        # Blocking save.
        async_handler_save_fn = asyncio_utils.as_async_function(handler.save)
        future.CommitFutureAwaitingContractedSignals(
            async_handler_save_fn(item_directory.get(), args=arg)
        ).result()
      item_end_time = time.time() - item_start_time
      time_taken_per_item[item_name] = f'{item_end_time:.8f}'
      time_taken_all_items += item_end_time

    commit_futures.extend(
        jax.tree.flatten(await asyncio.gather(*save_ops))[0] or []
    )
    end_time = time.time()
    logging.info(
        '[process=%s][thread=%s] Initiated'
        ' CompositeCheckpointHandler.async_save. Time taken: %fs'
        ' (all_items=%fs, per_item=%s, temp_paths=%f)',
        multihost.process_index(),
        threading.current_thread().name,
        end_time - start_time,
        time_taken_all_items,
        time_taken_per_item,
        end_time - start_time - time_taken_all_items,
    )
    return commit_futures

  def save(self, *args, **kwargs):
    """Saves synchronously."""

    async def async_save():
      commit_futures = await self.async_save(*args, **kwargs)
      if commit_futures:
        for f in commit_futures:
          f.result()

    asyncio_utils.run_sync(async_save())

  def _items_exist(
      self,
      directory: epath.Path,
      item_names: List[str],
  ) -> Mapping[str, bool]:
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=_CONCURRENT_WORKERS
    ) as executor:
      items_exist = executor.map(
          lambda item_name: self._get_item_directory(
              directory, item_name
          ).exists(),
          item_names,
      )
    return {
        item_name: exists for item_name, exists in zip(item_names, items_exist)
    }

  def _existing_items(self, directory: epath.Path) -> List[str]:
    return [p.name for p in directory.iterdir() if p.is_dir()]

  def restore(
      self,
      directory: epath.Path,
      args: Optional[CompositeArgs] = None,
  ) -> CompositeResults:
    """Restores the provided item synchronously.

    Restoration can happen in a variety of modes, depending on which `args` is
    passed:
      - `args` is not None and not empty. Item names present in `args` will be
      restored as long as they exist in the checkpoint and have a registered
      handler.
      - `args` is None or empty. All items in the checkpoint that have a
      registered handler will be restored.

    Args:
      directory: Path to restore from.
      args: CompositeArgs object used to restore individual sub-items. May be
        None or "empty" (`CompositeArgs()`), which is used to indicate that the
        handler should restore everything. If an individual item was not
        specified during the save, `None` will be returned for that item's
        entry.

    Returns:
      A CompositeResults object with keys matching `CompositeArgs`, or with keys
      for all known items as specified at creation.

    Raises:
      KeyError: If an item could not be restored.
    """

    items_to_handlers = dict(
        self._get_all_registered_and_unregistered_items_and_handlers()
    )
    existing_items = self._existing_items(directory)

    if args is None or not args.items():
      composite_args_items = {}
      item_handlers = None
      for item_name in existing_items:
        # Skip reserved items (they were not specifically requested).
        if item_name in RESERVED_ITEM_NAMES:
          continue

        if item_name not in items_to_handlers:
          item_handlers = (
              item_handlers or self.metadata(directory).item_handlers
          )
          try:
            # TODO(b/392622978): Support custom (local) handler registry.
            handler = handler_type_registry.get_handler_type(
                item_handlers[item_name]
            )
          except KeyError as e:
            raise ValueError(
                f'Item "{item_name}" was found in the checkpoint, but could not'
                ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
                ' call `restore` with an appropriate `CheckpointArgs` subclass.'
            ) from e
        else:
          handler = items_to_handlers[item_name]
        if handler is None:
          raise ValueError(
              f'Item with name: "{item_name}" had an undetermined'
              ' `CheckpointHandler` when restoring. Please ensure the handler'
              ' was specified during initialization, or use the appropriate'
              ' `CheckpointArgs` subclass to indicate the item type.'
          )
        _, restore_ckpt_args_cls = checkpoint_args.get_registered_args_cls(
            handler
        )
        # Try default constructing the args.
        try:
          composite_args_items[item_name] = restore_ckpt_args_cls()
        except TypeError as e:
          raise ValueError(
              'Attempted to construct default restoration arguments for item'
              f' "{item_name}", but the `CheckpointArgs` class of type'
              f' {restore_ckpt_args_cls} is not default constructible. Original'
              f' error: {e}.'
          ) from e
      args = CompositeArgs(**composite_args_items)

    restored = {}
    # Sort keys to maintain consistent ordering across processes, otherwise
    # we may hit timeouts if processes wait at different barriers in per-item
    # handlers.
    # TODO(b/295899152): Find a less brittle solution or support
    # async-compatible barrier function.
    for item_name in sorted(args.keys()):
      arg = args[item_name]
      if item_name not in existing_items:
        raise KeyError(
            f'Item "{item_name}" was not found in the checkpoint. Available'
            f' items: {existing_items}'
        )
      handler = self._get_or_set_handler(item_name, arg)
      restored[item_name] = handler.restore(
          self._get_item_directory(directory, item_name), args=arg
      )
    return CompositeResults(**restored)

  def _get_item_handlers(
      self,
      saved_metadata: checkpoint.StepMetadata,
      item_names: list[str],
      items_to_handlers: dict[str, CheckpointHandler | None],
  ) -> dict[str, checkpoint.CheckpointHandlerTypeStr | None]:
    if saved_metadata.item_handlers is not None:
      if not isinstance(saved_metadata.item_handlers, dict):
        raise ValueError(
            'StepMetadata.item_handlers must be a dict, but is a'
            f' {type(saved_metadata.item_handlers)}. This likely indicates that'
            ' the checkpoint was saved using a non-composite handler.'
        )
      # Keep relevant non-None handler typestrs from on-disk metadata.
      item_handlers: dict[str, checkpoint.CheckpointHandlerTypeStr] = (
          saved_metadata.item_handlers
      )
    else:
      item_handlers: dict[str, checkpoint.CheckpointHandlerTypeStr] = {}
    for item_name in item_names:
      if items_to_handlers.get(item_name) is None:
        logging.warning(
            'Item "%s" was found in the checkpoint, but could not'
            ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
            ' call `restore` with an appropriate `CheckpointArgs` subclass.',
            item_name,
        )
        # Don't overwrite if it was already set on disk.
        if item_name not in item_handlers:
          item_handlers[item_name] = None
        continue

      handler = items_to_handlers[item_name]
      assert handler is not None
      # If already set on disk (to non-None), don't overwrite.
      if item_handlers.get(item_name) is None:
        item_handlers[item_name] = handler.typestr()

    return item_handlers

  def _get_item_metadata(
      self,
      directory: epath.Path,
      item_names: list[str],
      items_to_handlers: dict[str, CheckpointHandler | None],
  ) -> dict[str, Any]:
    # item_metadata is not saved in StepMetadata, so we don't need to worry
    # about reading it from disk.
    item_metadata = {}
    for item_name in item_names:
      if items_to_handlers.get(item_name) is None:
        logging.warning(
            'Item "%s" was found in the checkpoint, but could not'
            ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
            ' call `restore` with an appropriate `CheckpointArgs` subclass.',
            item_name,
        )
        item_metadata[item_name] = None
        continue

      handler = items_to_handlers[item_name]
      assert handler is not None
      item_metadata[item_name] = handler.metadata(
          self._get_item_directory(directory, item_name)
      )
    return item_metadata

  def _get_metadata_base(
      self,
      directory: epath.Path,
      item_names: list[str],
      get_item_metadata: bool = True,
  ) -> StepMetadata:
    """Base implementation for metadata handling.

    Args:
      directory: Path to the checkpoint.
      item_names: List of item names to process.
      get_item_metadata: whether to get item metadata,

    Returns:
        A tuple of item handlers and item metadata.
    """
    items_to_handlers = dict(
        self._get_all_registered_and_unregistered_items_and_handlers()
    )

    serialized_metadata = self._metadata_store.read(
        checkpoint.step_metadata_file_path(directory)
    )
    saved_metadata = step_metadata_serialization.deserialize(
        serialized_metadata or {}
    )

    if get_item_metadata:
      item_metadata = CompositeItemMetadata(
          **self._get_item_metadata(directory, item_names, items_to_handlers)
      )
    else:
      item_metadata = None

    return dataclasses.replace(
        saved_metadata,
        item_handlers=self._get_item_handlers(
            saved_metadata, item_names, items_to_handlers
        ),
        item_metadata=item_metadata,
    )

  def metadata_from_temporary_paths(
      self, directory: epath.Path
  ) -> StepMetadata:
    """Metadata for each item in the temporary checkpoint."""
    if not directory.exists():
      raise FileNotFoundError(f'Directory does not exist: {directory}')

    return self._get_metadata_base(
        directory=directory,
        item_names=list(self._current_temporary_paths.keys()),
        get_item_metadata=False,
    )

  def metadata(self, directory: epath.Path) -> StepMetadata:
    """Metadata for each item in the checkpoint.

    This has much the same logic as `restore`, in the sense that it tries to
    restore the metadata for each item present in the checkpoint. However, for
    any items that do not have a registered handler, the metadata for that item
    will simply be returned as `None`.

    Args:
      directory: Path to the checkpoint.

    Returns:
      StepMetadata

    Raises:
      FileNotFoundError: If the directory does not exist.
    """
    if not directory.exists():
      raise FileNotFoundError(f'Directory does not exist: {directory}')

    try:
      existing_items = self._existing_items(directory)
    except OSError:
      existing_items = []
      logging.warning(
          'Failed to get existing items from directory %s. Will use items '
          'provided during initialization.',
          directory,
      )

    for item_tmp_dir in self._current_temporary_paths.values():
      tmp_dir_name = item_tmp_dir.get().name
      if tmp_dir_name in existing_items:
        raise ValueError(
            f'Item "{tmp_dir_name}" was found in the checkpoint, but it is a '
            'temporary item. Please call finalize() before metadata().'
        )

    return self._get_metadata_base(
        directory=directory,
        item_names=existing_items,
    )

  def finalize(self, directory: epath.Path):
    if not self._current_temporary_paths:
      raise ValueError('finalize() called before any items were saved.')
    for item_name, handler in _get_unique_registered_items_and_handlers(
        self._handler_registry
    ):
      tmp_dir = self._current_temporary_paths.get(item_name, None)
      if tmp_dir is None or handler is None:
        # Not an error, as some items may not have been saved.
        continue
      handler.finalize(tmp_dir.get())
      tmp_dir.finalize(
      )

      # Remove the temporary path once it has been finalized.
      self._current_temporary_paths.pop(item_name)

  def close(self):
    for _, handler in _get_unique_registered_items_and_handlers(
        self._handler_registry
    ):
      if handler is not None:
        handler.close()


@register_with_handler(
    CompositeCheckpointHandler, for_save=True, for_restore=True
)
class CompositeArgs(Composite, CheckpointArgs):
  """Args for wrapping multiple checkpoint items together."""

  ...


# Returned object of CompositeCheckpointHandler is an alias of CompositeArgs.
CompositeResults = CompositeArgs



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base test for Checkpointer."""

# pylint: disable=missing-class-docstring,protected-access,missing-function-docstring

import datetime
import threading
from typing import Any, Sequence
from unittest import mock

from absl.testing import parameterized
from etils import epath
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.serialization import serialization
import orbax.checkpoint.experimental.v1 as ocp
from orbax.checkpoint.experimental.v1._src.path import step as path_step_lib
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.testing import array_utils as array_test_utils
from orbax.checkpoint.experimental.v1._src.testing import handler_utils
from orbax.checkpoint.experimental.v1._src.testing import tree_utils as tree_test_utils
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types

Checkpointer = ocp.training.Checkpointer
save_decision_policies = ocp.training.save_decision_policies
preservation_policies = ocp.training.preservation_policies

RootMetadata = ocp.training.RootMetadata
CheckpointMetadata = ocp.training.CheckpointMetadata

Foo = handler_utils.Foo
Bar = handler_utils.Bar
Baz = handler_utils.Baz
AbstractFoo = handler_utils.AbstractFoo
AbstractBar = handler_utils.AbstractBar
AbstractBaz = handler_utils.AbstractBaz

ocp.handlers.register_handler(handler_utils.BazHandler)


class CheckpointerTestBase:

  class Test(parameterized.TestCase):
    """Test class."""

    def setUp(self):
      super().setUp()

      pytree, abstract_pytree = array_test_utils.create_sharded_pytree()
      numpy_pytree, abstract_numpy_pytree = (
          array_test_utils.create_numpy_pytree()
      )
      self.pytree = {
          'jax_array': pytree,
          'numpy_array': numpy_pytree,
      }
      self.abstract_pytree = {
          'jax_array': abstract_pytree,
          'numpy_array': abstract_numpy_pytree,
      }

      self.directory = epath.Path(
          self.create_tempdir(name='checkpointing_test').full_path
      )

      test_utils.set_tensorstore_driver_for_test()
      test_utils.sync_global_processes('CheckpointerTest:setup_complete')

    def tearDown(self):
      test_utils.sync_global_processes('CheckpointerTest:tests_complete')
      super().tearDown()

    def save_pytree(
        self,
        checkpointer: Checkpointer,
        step: int,
        pytree: tree_types.PyTreeOf[tree_types.LeafType],
        metrics: tree_types.JsonType | None = None,
        custom_metadata: tree_types.JsonType | None = None,
    ) -> bool:
      """Saves pytree with v0 CheckpointManager or v1 Checkpointer."""
      raise NotImplementedError()

    def save_checkpointables(
        self,
        checkpointer: Checkpointer,
        step: int,
        checkpointables: dict[str, Any],
        metrics: tree_types.JsonType | None = None,
        custom_metadata: tree_types.JsonType | None = None,
    ) -> bool:
      """Saves checkpointables with v0 CheckpointManager or v1 Checkpointer."""
      raise NotImplementedError()

    def test_properties(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.assertEqual(checkpointer.directory, self.directory)
      self.assertEmpty(checkpointer.checkpoints)
      self.assertIsNone(checkpointer.latest)

    def test_save_restore_pytree(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)

      with self.subTest('with_abstract_pytree'):
        loaded = checkpointer.load_pytree(0, self.abstract_pytree)
        test_utils.assert_tree_equal(self, self.pytree, loaded)

      with self.subTest('without_abstract_pytree'):
        loaded = checkpointer.load_pytree(0)
        test_utils.assert_tree_equal(self, self.pytree, loaded)

    @parameterized.parameters((True,), (False,))
    def test_load_latest_pytree(self, latest_arg_is_none):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree, metrics={'loss': 0.5})
      new_pytree = {
          'jax_array': self.pytree['jax_array'],
          'numpy_array': array_test_utils.create_numpy_pytree(add=1)[0],
      }
      self.save_pytree(checkpointer, 1, new_pytree, metrics={'loss': 0.4})

      latest = checkpointer.latest
      self.assertIsInstance(latest, ocp.training.CheckpointMetadata)
      self.assertEqual(latest.step, 1)
      self.assertDictEqual(latest.metrics, {'loss': 0.4})

      if latest_arg_is_none:
        loaded = checkpointer.load_pytree(abstract_pytree=self.abstract_pytree)
      else:
        loaded = checkpointer.load_pytree(latest, self.abstract_pytree)

      test_utils.assert_tree_equal(self, new_pytree, loaded)

    def test_overwrites(self):
      plus_one_pytree, _ = array_test_utils.create_numpy_pytree(add=1)
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)
      checkpointer.save_pytree(0, plus_one_pytree, overwrite=True)
      test_utils.assert_tree_equal(
          self, plus_one_pytree, checkpointer.load_pytree(0)
      )

    def test_step_already_exists(self):
      plus_one_pytree, _ = array_test_utils.create_numpy_pytree(add=1)
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)
      with self.assertRaises(ocp.training.errors.StepAlreadyExistsError):
        checkpointer.save_pytree(0, plus_one_pytree)

    def test_load_non_existent_step(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)
      with self.subTest('load_pytree'):
        with self.assertRaises(FileNotFoundError):
          checkpointer.load_pytree(1)
      with self.subTest('pytree_metadata'):
        with self.assertRaises(FileNotFoundError):
          checkpointer.pytree_metadata(1)

    @parameterized.parameters(
        (save_decision_policies.ContinuousCheckpointingPolicy(), range(10)),
        (save_decision_policies.FixedIntervalPolicy(1), range(10)),
        (save_decision_policies.FixedIntervalPolicy(2), range(0, 10, 2)),
        (
            save_decision_policies.AnySavePolicy([
                save_decision_policies.SpecificStepsPolicy((2, 3, 6)),
                save_decision_policies.InitialSavePolicy(),
            ]),
            [0, 2, 3, 6],
        ),
    )
    def test_steps(
        self,
        policy: save_decision_policies.SaveDecisionPolicy,
        expected_steps: Sequence[int],
    ):
      num_steps = 10
      checkpointer = Checkpointer(self.directory, save_decision_policy=policy)
      self.enter_context(checkpointer)
      for step in range(num_steps):
        self.assertEqual(checkpointer.should_save(step), step in expected_steps)
        saved = self.save_pytree(checkpointer, step, self.pytree)
        self.assertEqual(saved, step in expected_steps)

      self.assertLen(checkpointer.checkpoints, len(expected_steps))
      self.assertSequenceEqual(
          [c.step for c in checkpointer.checkpoints], expected_steps
      )
      assert checkpointer.latest is not None
      self.assertEqual(checkpointer.latest.step, expected_steps[-1])

    def test_force_save_ignores_save_decision_policy(self):
      checkpointer = Checkpointer(
          self.directory,
          save_decision_policy=save_decision_policies.FixedIntervalPolicy(2),
      )
      self.enter_context(checkpointer)

      self.assertTrue(checkpointer.save_pytree(0, self.pytree))
      self.assertFalse(checkpointer.save_pytree(1, self.pytree))
      self.assertTrue(checkpointer.save_pytree(2, self.pytree))

      self.assertLen(checkpointer.checkpoints, 2)
      self.assertSequenceEqual(
          [c.step for c in checkpointer.checkpoints], [0, 2]
      )

      self.assertTrue(checkpointer.save_pytree(3, self.pytree, force=True))
      self.assertTrue(checkpointer.save_pytree(4, self.pytree, force=True))
      self.assertTrue(checkpointer.save_pytree(5, self.pytree, force=True))

      self.assertLen(checkpointer.checkpoints, 5)
      self.assertSequenceEqual(
          [c.step for c in checkpointer.checkpoints], [0, 2, 3, 4, 5]
      )

    def test_garbage_collection(self):
      self.skipTest('TODO(cpgaffney): Implement.')

    def test_reload(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)
      self.save_pytree(checkpointer, 1, self.pytree)
      self.assertLen(checkpointer.checkpoints, 2)
      assert checkpointer.latest is not None
      self.assertEqual(checkpointer.latest.step, 1)

      (self.directory / '1').rmtree(missing_ok=True)

      # Properties still reflect cached state before reload.
      self.assertLen(checkpointer.checkpoints, 2)
      assert checkpointer.latest is not None
      self.assertEqual(checkpointer.latest.step, 1)

      checkpointer.reload()

      self.assertLen(checkpointer.checkpoints, 1)
      assert checkpointer.latest is not None
      self.assertEqual(checkpointer.latest.step, 0)

    def test_skips_when_ongoing_save(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      saved_0 = checkpointer.save_pytree_async(0, self.pytree)
      saved_1 = checkpointer.save_pytree(1, self.pytree)
      self.assertTrue(saved_0.result())
      self.assertFalse(saved_1)
      self.assertLen(checkpointer.checkpoints, 1)
      assert checkpointer.latest is not None
      self.assertEqual(checkpointer.latest.step, 0)

    def test_save_pytree_async(self):
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)

      start_serialize = threading.Event()
      original_serialize = serialization.async_serialize_from_host

      def mock_serialize(*args, **kwargs):
        start_serialize.wait()  # Wait for explicit signal before proceeding.
        return original_serialize(*args, **kwargs)

      # Serialization to disk does not start until receiving an explicit signal.
      self.enter_context(
          mock.patch.object(
              serialization, 'async_serialize_from_host', new=mock_serialize
          )
      )

      step = 0
      response = checkpointer.save_pytree_async(step, self.pytree)
      initial_d_files_mtimes = tree_test_utils.get_d_files_mtimes(
          self.directory / str(step)
      )
      self.assertFalse(
          tree_test_utils.is_pytree_checkpoint_complete(
              self.directory / str(step)
          )
      )
      start_serialize.set()

      response.result()
      final_d_files_mtimes = tree_test_utils.get_d_files_mtimes(
          self.directory / str(step)
      )
      self.assertNotEmpty(final_d_files_mtimes)
      self.assertNotEqual(initial_d_files_mtimes, final_d_files_mtimes)
      self.assertTrue(
          tree_test_utils.is_pytree_checkpoint_complete(
              self.directory / str(step)
          )
      )

    def test_close(self):
      checkpointer = Checkpointer(self.directory)
      step_path = self.directory / '0'
      checkpointer.save_pytree_async(0, self.pytree)
      self.assertFalse(step_path.exists())  # Not finalized yet.
      # But a tmp dir should have been created.
      self.assertNotEmpty(list(self.directory.iterdir()))
      checkpointer.close()
      self.assertTrue(step_path.exists())

    def test_context_manager_close(self):
      step_path = self.directory / '0'
      with Checkpointer(self.directory) as checkpointer:
        checkpointer.save_pytree_async(0, self.pytree)
        self.assertFalse(step_path.exists())  # Not finalized yet.
        # But a tmp dir should have been created.
        self.assertNotEmpty(list(self.directory.iterdir()))
      self.assertTrue(step_path.exists())

    def test_step_name_format(self):
      checkpointer = Checkpointer(
          self.directory,
          step_name_format=path_step_lib.standard_name_format(
              step_prefix='foo'
          ),
      )
      self.enter_context(checkpointer)
      self.save_pytree(checkpointer, 0, self.pytree)
      self.assertTrue((self.directory / 'foo_0').exists())
      self.assertFalse((self.directory / '0').exists())

    @parameterized.product(
        reinitialize_checkpointer=(True, False),
    )
    def test_root_metadata(self, reinitialize_checkpointer):
      checkpointer = Checkpointer(
          self.directory, custom_metadata={'foo': 'bar'}
      )
      self.enter_context(checkpointer)
      if reinitialize_checkpointer:
        # Does not overwrite custom metadata.
        Checkpointer(self.directory, custom_metadata={'baz': 2})
      root_metadata = checkpointer.root_metadata()
      self.assertIsInstance(root_metadata, RootMetadata)
      self.assertDictEqual(root_metadata.custom_metadata, {'foo': 'bar'})

    @parameterized.product(
        reinitialize_checkpointer=(True, False),
    )
    def test_pytree_metadata(self, reinitialize_checkpointer):
      checkpointer = Checkpointer(self.directory)
      self.save_pytree(
          checkpointer,
          0,
          self.pytree,
          metrics={'loss': 0.5},
          custom_metadata={'baz': 'qux'},
      )
      if reinitialize_checkpointer:
        checkpointer.close()
        checkpointer = Checkpointer(self.directory)
        self.enter_context(checkpointer)
      checkpoint_metadata = checkpointer.pytree_metadata(0)
      self.assertIsInstance(checkpoint_metadata, CheckpointMetadata)
      self.assertDictEqual(checkpoint_metadata.custom_metadata, {'baz': 'qux'})
      self.assertDictEqual(checkpoint_metadata.metrics, {'loss': 0.5})
      self.assertIsNotNone(checkpoint_metadata.init_timestamp_nsecs)
      self.assertIsNotNone(checkpoint_metadata.commit_timestamp_nsecs)
      self.assertIsInstance(checkpoint_metadata.metadata, dict)
      self.assertSameElements(
          checkpoint_metadata.metadata.keys(), ['jax_array', 'numpy_array']
      )

    @parameterized.product(
        reinitialize_checkpointer=(True, False),
    )
    def test_checkpointables_metadata(self, reinitialize_checkpointer):
      checkpointer = Checkpointer(self.directory)
      self.save_checkpointables(
          checkpointer,
          0,
          {'pytree': self.pytree, 'baz': Baz(123, 'hi')},
          metrics={'loss': 0.5},
          custom_metadata={'baz': 'qux'},
      )
      if reinitialize_checkpointer:
        checkpointer.close()
        checkpointer = Checkpointer(self.directory)
        self.enter_context(checkpointer)
      checkpoint_metadata = checkpointer.checkpointables_metadata(0)
      self.assertIsInstance(checkpoint_metadata, CheckpointMetadata)
      self.assertDictEqual(checkpoint_metadata.custom_metadata, {'baz': 'qux'})
      self.assertDictEqual(checkpoint_metadata.metrics, {'loss': 0.5})
      self.assertIsNotNone(checkpoint_metadata.init_timestamp_nsecs)
      self.assertIsNotNone(checkpoint_metadata.commit_timestamp_nsecs)
      self.assertIsInstance(checkpoint_metadata.metadata, dict)
      self.assertSameElements(
          checkpoint_metadata.metadata.keys(), ['pytree', 'baz']
      )
      self.assertSameElements(
          checkpoint_metadata.metadata['pytree'].keys(),
          ['jax_array', 'numpy_array'],
      )
      self.assertIsInstance(checkpoint_metadata.metadata['baz'], AbstractBaz)

    def test_custom_checkpointables(self):
      """Test custom checkpointables are saved and loaded.

      Subclasses must call it within custom `checkpointables_options` Context.
      """
      checkpointables = {
          'pytree': self.pytree,
          'foo': Foo(123, 'hi'),
          'bar': Bar(456, 'bye'),
      }
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_checkpointables(checkpointer, 0, checkpointables)

      with self.subTest('load'):
        loaded = checkpointer.load_checkpointables(0)
        self.assertSameElements(loaded.keys(), ['pytree', 'foo', 'bar'])
        test_utils.assert_tree_equal(
            self, checkpointables['pytree'], loaded['pytree']
        )
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_with_free_function'):
        loaded = ocp.load_checkpointables(self.directory / '0')
        self.assertSameElements(loaded.keys(), ['pytree', 'foo', 'bar'])
        test_utils.assert_tree_equal(
            self, checkpointables['pytree'], loaded['pytree']
        )
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_with_abstract_checkpointables'):
        abstract_checkpointables = {
            'pytree': self.abstract_pytree,
            'foo': AbstractFoo(),
            'bar': AbstractBar(),
        }
        loaded = checkpointer.load_checkpointables(0, abstract_checkpointables)
        self.assertSameElements(loaded.keys(), ['pytree', 'foo', 'bar'])
        test_utils.assert_tree_equal(self, self.pytree, loaded['pytree'])
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_with_abstract_checkpointables_none_values'):
        abstract_checkpointables = {
            'pytree': None,
            'foo': None,
            'bar': None,
        }
        loaded = checkpointer.load_checkpointables(0, abstract_checkpointables)
        self.assertSameElements(loaded.keys(), ['pytree', 'foo', 'bar'])
        test_utils.assert_tree_equal(
            self, checkpointables['pytree'], loaded['pytree']
        )
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_partial'):
        abstract_checkpointables = {
            'foo': None,
        }
        loaded = checkpointer.load_checkpointables(0, abstract_checkpointables)
        self.assertSameElements(loaded.keys(), ['foo'])
        self.assertEqual(checkpointables['foo'], loaded['foo'])

    def test_load_with_switched_abstract_checkpointables(self):
      """Test load with switched abstract checkpointables.

      Subclasses must call it within custom `checkpointables_options` Context.
      """
      checkpointables = {
          'pytree': self.pytree,
          'foo': Foo(123, 'hi'),
          'bar': Bar(456, 'bye'),
      }
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_checkpointables(checkpointer, 0, checkpointables)

      abstract_checkpointables = {
          'foo': AbstractBar(),
          'bar': AbstractFoo(),
      }
      loaded = checkpointer.load_checkpointables(0, abstract_checkpointables)
      self.assertSameElements(loaded.keys(), ['foo', 'bar'])
      self.assertEqual(Bar(123, 'hi'), loaded['foo'])
      self.assertEqual(Foo(456, 'bye'), loaded['bar'])

    def test_different_custom_checkpointables(self):
      """Test different custom checkpointables are saved and loaded.

      Subclasses must call it within custom `checkpointables_options` Context.
      """
      checkpointer = Checkpointer(self.directory)
      self.enter_context(checkpointer)
      self.save_checkpointables(checkpointer, 0, {'foo': Foo(123, 'hi')})
      checkpointer.save_checkpointables(1, {'bar': Bar(456, 'bye')})

      loaded = checkpointer.load_checkpointables(0)
      self.assertSameElements(loaded.keys(), ['foo'])
      self.assertEqual(Foo(123, 'hi'), loaded['foo'])

      loaded = checkpointer.load_checkpointables(1)
      self.assertSameElements(loaded.keys(), ['bar'])
      self.assertEqual(Bar(456, 'bye'), loaded['bar'])

    def test_custom_save_decision_policy(self):
      save_delta = datetime.timedelta(seconds=0.03)

      class ArbitrarySavePolicy(save_decision_policies.SaveDecisionPolicy):

        def should_save(
            self,
            step: CheckpointMetadata,
            previous_steps: Sequence[CheckpointMetadata],
            *,
            context: save_decision_policies.DecisionContext,
        ) -> bool:
          save_result = False
          is_primary_host = multihost.is_primary_host(
              context.multiprocessing_options.primary_host
          )
          if is_primary_host:
            if not previous_steps:
              save_result = True
            else:
              time_delta = step.time - previous_steps[-1].time
              save_result = step.step % 2 == 0 and time_delta >= save_delta
          save_result = bool(
              multihost.broadcast_one_to_all(
                  save_result, is_source=is_primary_host
              )
          )
          return save_result

      checkpointer = Checkpointer(
          self.directory, save_decision_policy=ArbitrarySavePolicy()
      )
      self.enter_context(checkpointer)
      for step in range(0, 30):
        self.save_pytree(checkpointer, step, self.pytree)

      self.assertNotEmpty(checkpointer.checkpoints)
      self.assertLess(len(checkpointer.checkpoints), 30)
      checkpointer_metadata = [
          checkpointer.pytree_metadata(metadata.step)
          for metadata in checkpointer.checkpoints
      ]
      for i in range(1, len(checkpointer.checkpoints)):
        time_delta = (
            checkpointer_metadata[i].time - checkpointer_metadata[i - 1].time
        )
        self.assertGreaterEqual(time_delta, save_delta)

    @parameterized.parameters(
        (None, range(10)),
        (ocp.training.preservation_policies.PreserveAll(), range(10)),
        (preservation_policies.LatestN(3), range(7, 10)),
        (preservation_policies.EveryNSteps(4), [0, 4, 8]),
        (preservation_policies.EveryNSeconds(40), range(0, 10, 2)),
        (
            preservation_policies.AnyPreservationPolicy([
                preservation_policies.LatestN(3),
                preservation_policies.CustomSteps([1, 3, 9]),
            ]),
            [1, 3, 7, 8, 9],
        ),
    )
    def test_preservation(self, policy, expected_steps):
      num_steps = 10
      checkpointer = Checkpointer(self.directory, preservation_policy=policy)
      for step in range(num_steps):

        class CustomDatetime(datetime.datetime):

          @classmethod
          def now(cls, tz=None):
            return datetime.datetime.fromtimestamp(
                step * 20, tz=tz  # pylint: disable=cell-var-from-loop
            )

        with mock.patch('datetime.datetime', new=CustomDatetime):
          # mock_dt.now.return_value = datetime.datetime.fromtimestamp(
          #     checkpoint_times[step]
          # )
          # mock_dt.fromtimestamp.side_effect = datetime.datetime.fromtimestamp
          # mock_dt.timestamp.return_value = checkpoint_times[step]
          checkpointer.save_pytree(step, self.pytree)

      self.assertLen(checkpointer.checkpoints, len(expected_steps))
      self.assertSequenceEqual(
          [c.step for c in checkpointer.checkpoints], expected_steps
      )
      self.assertSequenceEqual(
          [c.time for c in checkpointer.checkpoints],
          [
              datetime.datetime.fromtimestamp(
                  step * 20, tz=datetime.timezone.utc
              )
              for step in expected_steps
          ],
      )
      checkpointer.close()

    @parameterized.parameters(
        (
            preservation_policies.BestN(
                get_metric_fn=lambda m: m['accuracy'], n=2
            ),
            [0, 4],
        ),
        (
            preservation_policies.BestN(
                get_metric_fn=lambda m: m['loss'], reverse=True, n=2
            ),
            [0, 1],
        ),
        (
            preservation_policies.BestN(
                get_metric_fn=lambda m: m['accuracy'], n=None
            ),
            range(5),
        ),
        (
            preservation_policies.BestN(
                get_metric_fn=lambda m: m['accuracy'], n=0
            ),
            [],
        ),
    )
    def test_preservation_metrics(self, policy, expected_steps):
      num_steps = 5
      all_metrics = [
          {'loss': 0.3, 'accuracy': 0.8},
          {'loss': 0.2, 'accuracy': 0.6},
          {'loss': 0.8, 'accuracy': 0.2},
          {'loss': 0.9, 'accuracy': 0.3},
          {'loss': 0.4, 'accuracy': 0.9},
      ]
      checkpointer = Checkpointer(self.directory, preservation_policy=policy)
      for step in range(num_steps):
        checkpointer.save_pytree(step, self.pytree, metrics=all_metrics[step])

      self.assertLen(checkpointer.checkpoints, len(expected_steps))
      self.assertSequenceEqual(
          [c.step for c in checkpointer.checkpoints], expected_steps
      )
      self.assertSequenceEqual(
          [c.metrics for c in checkpointer.checkpoints],
          [all_metrics[step] for step in expected_steps],
      )
      checkpointer.close()
